from .base import (
    AllowAny,
    BasePermission,
    DenyAll,
    IsAdminUser,
    IsAuthenticated,
    IsAuthenticatedOrReadOnly,
)

__all__ = [
    "AllowAny",
    "BasePermission",
    "DenyAll",
    "IsAdminUser",
    "IsAuthenticated",
    "IsAuthenticatedOrReadOnly",
]
